r"""
    Routines for calculating the Helmholtz energy for the SAFT-gamma equation of state.
    
    Equations referenced in this code are from V. Papaioannou et al. J. Chem. Phys. 140 054107 2014
    
"""

import numpy as np
import logging
#from profilehooks import profile

from . import constants

#@profile
def calc_a1s(rho, Cmol2seg, l_kl, zetax, epsilonkl, dkl):
    r""" 
    Return a1s,kl(rho*Cmol2seg,l_kl) in K as defined in eq. 25.
    
    Used in the calculation of :math:`A_1` the first order term of the perturbation expansion corresponding to the mean-attractive energy.

    Parameters
    ----------
    rho : numpy.ndarray
        Number density of system [molecules/m^3]
    Cmol2seg : float
        Conversion factor from from molecular number density, :math:`\rho`, to segment (i.e. group) number density, :math:`\rho_S`. Shown in eq. 13
    l_kl : numpy.ndarray
        Matrix of mie potential exponents for k,l groups
    zetax : numpy.ndarray
        Matrix of hypothetical packing fraction based on hard sphere diameter for groups (k,l)
    epsilonkl : numpy.ndarray
        Matrix of well depths for groups (k,l)
    dkl : numpy.ndarray
        Matrix of hard sphere diameters for groups (k,l)

    Returns
    -------
    a1s : numpy.ndarray
        Matrix used in the calculation of :math:`A_1` the first order term of the perturbation expansion corresponding to the mean-attractive energy, size is the Ngroups by Ngroups
    """

    #logger = logging.getLogger(__name__)

    nbeads = np.size(dkl, axis=0)
    zetax_pow = np.zeros((np.size(rho), 4))
    zetax_pow[:, 0] = zetax
    for i in range(1, 4):
        zetax_pow[:, i] = zetax_pow[:, i - 1] * zetax_pow[:, 0]

    # check if you have more than 1 bead types
    if np.size(np.shape(l_kl)) == 2:
        etakl = np.zeros((np.size(rho), nbeads, nbeads))
        for k in range(nbeads):
            for l in range(nbeads):
                cikl = np.inner(constants.ckl_coef, np.transpose(np.array([1.0, l_kl[k, l]**-1, l_kl[k, l]**-2, l_kl[k, l]**-3])))
                etakl[:, k, l] = np.einsum("ij,j", zetax_pow, cikl)
        a1s = np.einsum("ijk,jk->ijk", (1.0 - (etakl / 2.0)) / ((1.0 - etakl)**3),
                        -2.0 * np.pi * Cmol2seg * ((epsilonkl * (dkl**3)) / (l_kl - 3.0)))
        # a1s is 4D matrix
        a1s = np.einsum("i,ijk->ijk", rho, a1s)  # {BottleNeck}

    elif np.size(np.shape(l_kl)) == 1:
        etakl = np.zeros((np.size(rho), nbeads))
        for k in range(nbeads):
            cikl = np.inner(constants.ckl_coef, np.transpose(np.array([1.0, l_kl[k]**-1, l_kl[k]**-2, l_kl[k]**-3])))
            etakl[:, k] = np.einsum("ij,j", zetax_pow, cikl)
        a1s = np.einsum("ij,j->ij", (1.0 - (etakl / 2.0)) / ((1.0 - etakl)**3),
                        -2.0 * np.pi * Cmol2seg * ((epsilonkl * (dkl**3)) / (l_kl - 3.0)))
        # a1s is 3D matrix
        a1s = np.einsum("i,ij->ij", rho, a1s)
    else:
        print('Error in calc_a1s, unexpected array size')

    return a1s

#@profile
def calc_da1sii_drhos(rho, Cmol2seg, l_kl, zetax, epsilonkl, dkl):
    r""" 
    Return da1s,kl(rho*Cmol2seg,l_kl)/rhos in K.
    
    Used in the calculation of :math:`A_chain` the first order term of the perturbation expansion corresponding to the mean-attractive energy.

    Parameters
    ----------
    rho : numpy.ndarray
        Number density of system [molecules/m^3]
    Cmol2seg : float
        Conversion factor from from molecular number density, :math:`\rho`, to segment (i.e. group) number density, :math:`\rho_S`. Shown in eq. 13
    l_kl : numpy.ndarray
        Matrix of mie potential exponents for k,l groups
    zetax : numpy.ndarray
        Matrix of hypothetical packing fraction based on hard sphere diameter for groups (k,l)
    epsilonkl : numpy.ndarray
        Matrix of well depths for groups (k,l)
    dkl : numpy.ndarray
        Matrix of hard sphere diameters for groups (k,l)

    Returns
    -------
    da1s_drhos : numpy.ndarray
        Matrix used in the calculation of :math:`A_1` the first order term of the perturbation expansion corresponding to the mean-attractive energy, size is the Ngroups by Ngroups
    """

    #logger = logging.getLogger(__name__)

    nbeads = np.size(dkl, axis=0)
    zetax_pow = np.zeros((np.size(rho), 4))
    zetax_pow[:, 0] = zetax
    for i in range(1, 4):
        zetax_pow[:, i] = zetax_pow[:, i - 1] * zetax_pow[:, 0]

    # check if you have more than 1 bead types
    if np.size(np.shape(l_kl)) == 2:
        etakl = np.zeros((np.size(rho), nbeads, nbeads))
        rhos_detakl_drhos = np.zeros((np.size(rho), nbeads, nbeads))
        for k in range(nbeads):
            for l in range(nbeads):
                # Constants to calculate eta_eff
                cikl = np.inner(constants.ckl_coef, np.transpose(np.array([1.0, l_kl[k, l]**-1, l_kl[k, l]**-2, l_kl[k, l]**-3])))
                etakl[:, k, l] = np.einsum("ij,j", zetax_pow, cikl)
                rhos_detakl_drhos[:, k, l] = np.einsum("ij,j", zetax_pow, cikl*np.array([1.0,2.0,3.0,4.0]))
        da1s_drhos = np.einsum("ijk,jk->ijk", (1.0 - (etakl / 2.0)) / ((1.0 - etakl)**3) + (5.0-2.0*etakl)/(2.0*(1.0-etakl)**4)*rhos_detakl_drhos,
                        -2.0 * np.pi * ((epsilonkl * (dkl**3)) / (l_kl - 3.0)))

    elif np.size(np.shape(l_kl)) == 1:
        etakl = np.zeros((np.size(rho), nbeads))
        rhos_detakl_drhos = np.zeros((np.size(rho), nbeads))
        for k in range(nbeads):
            cikl = np.inner(constants.ckl_coef, np.transpose(np.array([1.0, l_kl[k]**-1, l_kl[k]**-2, l_kl[k]**-3])))
            etakl[:, k] = np.einsum("ij,j", zetax_pow, cikl)
            rhos_detakl_drhos[:, k] = np.einsum("ij,j", zetax_pow, cikl*np.array([1.,2.,3.,4.]))

        tmp1 = (1.0 - (etakl / 2.0)) / ((1.0 - etakl)**3) + (5.0-2.0*etakl)/(2.0*(1.0-etakl)**4)*rhos_detakl_drhos
        tmp2 = -2.0 * np.pi * ((epsilonkl * (dkl**3)) / (l_kl - 3.0))
        da1s_drhos = np.einsum("ij,j->ij",tmp1,tmp2)
#        da1s_drhos = np.einsum("ij,j->ij", (1.0 - (etakl / 2.0)) / ((1.0 - etakl)**3) + (5.0-2.0*etakl)/(2.0*(1.0-etakl)**4)*rhos_detakl_drhos, -2.0 * np.pi * ((epsilonkl * (dkl**3)) / (l_kl - 3.0)))
    else:
        print('Error in calc_da1s_drhos, unexpected array size')

    return da1s_drhos

#@profile
def calc_Xika(indices, rho, xi, nui, nk, Fklab, Kklab, Iij, maxiter=500, tol=1e-12, damp=.1):
    r""" 
    Calculate the fraction of molecules of component i that are not bonded at a site of type a on group k.

    Parameters
    ----------
    indices : list[list]
        A list of sets of (component, bead, site) to identify the values of the Xika matrix that are being fit
    rho : numpy.ndarray
        Number density of system [molecules/m^3]
    xi : numpy.ndarray
        Mole fraction of each component, sum(xi) should equal 1.0
    nui : numpy.array
        :math:`\nu_{i,k}/k_B`, Array of number of components by number of bead types. Defines the number of each type of group in each component. 
    nk : numpy.ndarray
        For each bead the number of each type of site
    delta : numpy.ndarray
        The association strength between a site of type a on a group of type k of component i and a site of type b on a group of type l of component j. eq. 66
    maxiter : int, Optional, default: 500
        Maximum number of iteration for minimization
    tol : float, Optional, default: 1e-12
        Once matrix converges.
    damp : float, Optional, default: 0.1
        Only add a fraction of the new matrix values to update the guess

    Returns
    -------
    Xika : numpy.ndarray
        NoteHere
    """

    #logger = logging.getLogger(__name__)

    nbeads    = nui.shape[1]
    ncomp     = np.size(xi)
    nsitesmax = np.size(nk, axis=1)
    nrho      = len(rho)
    l_ind = len(indices)


    Xika_final = np.ones((nrho,ncomp, nbeads, nsitesmax))
    err_array   = np.zeros(nrho)

    # Parallelize here, with respect to rho!
    Xika_elements = .5*np.ones(len(indices))
    for r in range(nrho):
        for knd in range(maxiter):

            Xika_elements_new = np.ones(len(Xika_elements))
            ind = 0
            for iind in range(l_ind):
                i, k, a = indices[iind]
                jnd = 0
                for jjnd in range(l_ind):
                    j, l, b = indices[jjnd]
                    delta = Fklab[k, l, a, b] * Kklab[k, l, a, b] * Iij[r,i, j]
                    Xika_elements_new[ind] += rho[r] * xi[j] * nui[j,l] * nk[l,b] *    Xika_elements[jnd] * delta
                    jnd += 1
                ind += 1
            Xika_elements_new = 1./Xika_elements_new
            obj = np.sum(np.abs(Xika_elements_new - Xika_elements))

            if obj < tol:
                break
            else:
                if obj/max(Xika_elements) > 1e+3:
                    Xika_elements = Xika_elements + damp*(Xika_elements_new - Xika_elements)
                else:
                    Xika_elements = Xika_elements_new

        err_array[r] = obj

        for jjnd in range(l_ind):
            i,k,a = indices[jjnd]
            Xika_final[r,i,k,a] = Xika_elements[jjnd]
    
    return Xika_final, err_array




"""
Unit and regression test for the despasito package.
"""

# Import package, test suite, and other packages as needed
import despasito.thermodynamics.calc as calc
import despasito.thermodynamics as thermo
import despasito.equations_of_state
import pytest
import sys
import numpy as np

Tlist = [323.2]
xilist = [[0.01,0.99]]
yilist = [[0.97702, 0.02298]]
Plist = [4492927.45]
rho_co2_h2o = np.array([19986.78358869])
P = np.array([4099056.163132072])

beads_co2_h2o = ['CO2', 'H2O']
nui_co2_h2o = np.array([[1., 0.],[0., 1.]])
beadlibrary_co2_h2o = {'CO2': {'epsilon': 207.89, 'l_a': 5.055, 'l_r': 26.408, 'sigma': 3.05e-10, 'Sk': 0.8468, 'Vks': 2, 'mass': 0.04401, 'NkH': 1, 'Nka1': 1},'H2O': {'epsilon': 266.68, 'l_a': 6.0, 'l_r': 17.02, 'sigma': 3.0063e-10, 'Sk': 1.0, 'Vks': 1, 'mass': 0.018015, 'NkH': 2, 'Nke1': 2, 'epsilonHe1': 1985.4, 'KHe1': 1.0169e-28}}
crosslibrary_co2_h2o = {'CO2': {'H2O': {'epsilon': 226.38, 'epsilonHe1': 2200.0, 'KHe1': 9.1419e-29}}}
sitenames_co2_h2o = ['H', 'e1', 'a1'] 
epsilonHB_co2_h2o = np.array([[[[   0.,     0.,     0. ], \
                       [   0.,     0.,     0. ], \
                       [   0.,     0.,     0. ]], \
                      [[   0.,  2200.,     0. ], \
                       [   0.,     0.,     0. ], \
                       [   0.,     0.,     0. ]]], \
                     [[[   0.,     0.,     0. ], \
                       [2200.,     0.,     0. ], \
                       [   0.,     0.,     0. ]], \
                      [[   0.,  1985.4,    0. ], \
                       [1985.4,    0.,     0. ], \
                       [   0.,     0.,     0. ]]]])
eos_co2_h2o = despasito.equations_of_state.eos(eos="saft.gamma_mie",beads=beads_co2_h2o,nui=nui_co2_h2o,beadlibrary=beadlibrary_co2_h2o,crosslibrary=crosslibrary_co2_h2o,sitenames=sitenames_co2_h2o)

def test_thermo_import():
#    """Sample test, will always pass so long as import statement worked"""
    assert "despasito.thermodynamics" in sys.modules

def test_phase_xiT(eos=eos_co2_h2o,Tlist=Tlist,xilist=xilist):

    output = thermo.thermo(eos,{"calculation_type":"phase_xiT","Tlist":Tlist,"xilist":xilist})
        
#    assert output["P"][0]==pytest.approx(1223211.573700886,abs=1e+1) and output["yi"][0]==pytest.approx([0.94880358, 0.05119642],abs=1e-4)
    assert output["P"][0]==pytest.approx(3042623.2,abs=1e+1) and output["yi"][0]==pytest.approx([0.97701902, 0.02298098],abs=1e-4)

#def test_phase_yiT(eos=eos_co2_h2o,Tlist=Tlist,yilist=yilist):
#    output = thermo.thermo(eos,{"calculation_type":"phase_yiT","Tlist":Tlist,"yilist":yilist})
#    assert output["P"][0]==pytest.approx(1174228.60,abs=1e+1) and output["xi"][0]==pytest.approx([0.04803023, 0.95196977],abs=1e-4)

def test_sat_props(eos=eos_co2_h2o,Tlist=Tlist):

    output = thermo.thermo(eos,{"calculation_type":"sat_props","Tlist":Tlist,"xilist":[np.array([0.0, 1.0])]})

    assert output["Psat"][0]==pytest.approx(12314.30,abs=1e+1) and output["rhol"][0]==pytest.approx(54700.25,abs=1e-1), output["rhol"][0]==pytest.approx(2371.38970066,abs=1e-1)

def test_liquid_properties(eos=eos_co2_h2o,Tlist=Tlist,xilist=xilist,Plist=Plist):

    output = thermo.thermo(eos,{"calculation_type":"liquid_properties","Tlist":Tlist,"Plist":Plist,"xilist":xilist})

#    assert output["rhol"][0]==pytest.approx(54072.87630577754,abs=1e-1) and output["phil"][0]==pytest.approx(np.array([2646.44010, 0.120295122]),abs=1e-1)
    assert output["rhol"][0]==pytest.approx(54156.297,abs=1e-1) and output["phil"][0]==pytest.approx(np.array([6.04140887e+01, 2.79514245e-03]),abs=1e-1)

def test_vapor_properties(eos=eos_co2_h2o,Tlist=Tlist,yilist=yilist,Plist=Plist):

    output = thermo.thermo(eos,{"calculation_type":"vapor_properties","Tlist":Tlist,"Plist":Plist,"yilist":yilist})

#    assert output["rhov"][0]==pytest.approx(37.85937201,abs=1e-1) and output["phiv"][0]==pytest.approx(np.array([2.45619145, 0.37836741]),abs=1e-1)
    assert output["rhov"][0]==pytest.approx(2156.81,abs=1e-1) and output["phiv"][0]==pytest.approx(np.array([0.90729601, 0.13974291]),abs=1e-1)
    



"""
Unit and regression test for the despasito package.
"""

# Import package, test suite, and other packages as needed
import despasito.equations_of_state.saft.gamma_mie
import despasito.equations_of_state.saft.gamma_mie_funcs

import despasito.equations_of_state.saft.solv_assoc as solv_assoc

import pytest
import sys
import numpy as np

xi_co2_ben = np.array([0.2, 0.2])
beads_co2_ben = ['CO2', 'benzene']
nui_co2_ben = np.array([[1., 0.],[0., 1.]])
beadlibrary_co2_ben = {'CO2': {'epsilon': 361.69, 'l_a': 6.66, 'l_r': 23.0, 'sigma': 3.741e-10, 'Sk': 1.0, 'Vks': 1.0, 'mass': 0.04401}, 'baCH': {'epsilon': 371.53, 'l_a': 6.0, 'l_r': 14.756, 'sigma': 4.0578e-10, 'Sk': 0.32184, 'Vks': 1.0, 'mass': 0.01302}, 'baCHCH': {'epsilon': 243.88, 'l_a': 6.0, 'l_r': 11.58, 'sigma': 3.482e-10, 'Sk': 1.0, 'Vks': 1.0, 'mass': 0.02604}, 'baCHCHCH': {'epsilon': 353.93, 'l_a': 6.0, 'l_r': 14.23, 'sigma': 3.978e-10, 'Sk': 1.0, 'Vks': 1.0, 'mass': 0.03905}, 'benzene': {'epsilon': 658.17, 'l_a': 6.0, 'l_r': 32.0, 'sigma': 3.842e-10, 'Sk': 1.0, 'Vks': 1.0, 'mass': 0.07811}}

xi_co2_h2o = np.array([0.78988277, 0.21011723])
beads_co2_h2o = ['CO2', 'H2O']
nui_co2_h2o = np.array([[1., 0.],[0., 1.]])
beadlibrary_co2_h2o = {'CO2': {'epsilon': 207.89, 'l_a': 5.055, 'l_r': 26.408, 'sigma': 3.05e-10, 'Sk': 0.8468, 'Vks': 2, 'mass': 0.04401, 'NkH': 1, 'Nka1': 1},'H2O': {'epsilon': 266.68, 'l_a': 6.0, 'l_r': 17.02, 'sigma': 3.0063e-10, 'Sk': 1.0, 'Vks': 1, 'mass': 0.018015, 'NkH': 2, 'Nke1': 2, 'epsilonHe1': 1985.4, 'KHe1': 1.0169e-28}}
crosslibrary_co2_h2o = {'CO2': {'H2O': {'epsilon': 226.38, 'epsilonHe1': 2200.0, 'KHe1': 9.1419e-29}}}
sitenames_co2_h2o = ['H', 'e1', 'a1'] 
epsilonHB_co2_h2o = np.array([[[[   0.,     0.,     0. ], \
                       [   0.,     0.,     0. ], \
                       [   0.,     0.,     0. ]], \
                      [[   0.,  2200.,     0. ], \
                       [   0.,     0.,     0. ], \
                       [   0.,     0.,     0. ]]], \
                     [[[   0.,     0.,     0. ], \
                       [2200.,     0.,     0. ], \
                       [   0.,     0.,     0. ]], \
                      [[   0.,  1985.4,    0. ], \
                       [1985.4,    0.,     0. ], \
                       [   0.,     0.,     0. ]]]])
eos_co2_h2o = despasito.equations_of_state.eos(eos="saft.gamma_mie",beads=beads_co2_h2o,nui=nui_co2_h2o,beadlibrary=beadlibrary_co2_h2o,crosslibrary=crosslibrary_co2_h2o,sitenames=sitenames_co2_h2o)
T = 323.2 
rho_co2_h2o = np.array([21146.16997993]) 
P = np.array([1713500.67089664])

def test_saft_gamma_mie_imported():
#    """Sample test, will always pass so long as import statement worked"""
    assert "despasito.equations_of_state.saft.gamma_mie" in sys.modules

def test_saft_gamma_mie_class_noassoc(beads=beads_co2_ben,nui=nui_co2_ben,beadlibrary=beadlibrary_co2_ben):    
#   """Test ability to create EOS object without association sites"""
    eos_class = despasito.equations_of_state.eos(eos="saft.gamma_mie",beads=beads,nui=nui,beadlibrary=beadlibrary)
    assert (eos_class._massi==np.array([0.04401, 0.07811])).all()

def test_saft_gamma_mie_class_assoc(beads=beads_co2_h2o,nui=nui_co2_h2o,beadlibrary=beadlibrary_co2_h2o,crosslibrary=crosslibrary_co2_h2o,sitenames=sitenames_co2_h2o,epsilonHB=epsilonHB_co2_h2o):
#   """Test ability to create EOS object with association sites"""
    eos_class = despasito.equations_of_state.eos(eos="saft.gamma_mie",beads=beads,nui=nui,beadlibrary=beadlibrary,crosslibrary=crosslibrary,sitenames=sitenames)
    assert (eos_class._epsilonHB==epsilonHB).all()

def test_saft_gamma_mie_class_assoc_P(T=T,xi=xi_co2_h2o,eos=eos_co2_h2o,rho=rho_co2_h2o):
#   """Test ability to predict P with association sites"""
    P = eos.P(rho,T,xi)[0]
    assert P == pytest.approx(14807941.78,abs=1e-1)

def test_saft_gamma_mie_class_assoc_mu(P=P,xi=xi_co2_h2o,T=T,eos=eos_co2_h2o,rho=rho_co2_h2o):
#   """Test ability to predict P with association sites"""
    phi = eos.fugacity_coefficient(P, rho, xi, T)
#    assert mui == pytest.approx(np.array([1.61884825, -4.09022886]),abs=1e-4)
    assert phi == pytest.approx(np.array([4.35827475, 0.02554506]),abs=1e-4)





"""
Process command line argparse and initiate logging settings.
"""

from .main import run, commandline_parser
import os
import logging
import logging.handlers

parser = commandline_parser()

## Extract arguments
quiet = False
args = parser.parse_args()

if args.verbose == 0:
    quiet = True
    args.verbose = 20
elif args.verbose < 4:
    args.verbose = (4-args.verbose)*10
else:
    args.verbose = 10

# Logging

logger = logging.getLogger()
logger.setLevel(args.verbose)

# Set up rotating log files
if os.path.isfile(args.logFile):
    os.remove(args.logFile)
log_file_handler = logging.handlers.RotatingFileHandler(args.logFile)
log_file_handler.setFormatter( logging.Formatter('%(asctime)s [%(levelname)s](%(name)s:%(funcName)s:%(lineno)d): %(message)s') )
log_file_handler.setLevel(args.verbose)
logger.addHandler(log_file_handler)

if quiet == False:
    # Set up logging to console
    console_handler = logging.StreamHandler() # sys.stderr
    console_handler.setFormatter( logging.Formatter('[%(levelname)s](%(name)s): %(message)s') )
    console_handler.setLevel(args.verbose)
    logger.addHandler(console_handler)

logging.info("Input args: {}".format(args))
logging.info("JIT compilation: {}".format(args.jit))

# Threads
# if args.threads != None:
#     threadcount = args.threads
# else:
#     threadcount = 1

# Run program
if args.input:
    kwargs = {"filename":args.input}
else:
    kwargs = {}
#kwargs["logFile"] = args.logFile
kwargs["threads"] = args.threads
kwargs["path"] = args.path
kwargs["jit" ] = args.jit

run(**kwargs)



r"""
Objects for storing and producing objective values for comparing experimental data to EOS predictions.    
"""

import numpy as np
import logging

from despasito.thermodynamics import thermo
from despasito.fit_parameters import fit_funcs as ff
from despasito.fit_parameters.interface import ExpDataTemplate

##################################################################
#                                                                #
#                       Liquid Density                           #
#                                                                #
##################################################################
class Data(ExpDataTemplate):

    r"""
    Object for liquid density data. This data is evaluated with "liquid_properties". 

    Parameters
    ----------
    data_dict : dict
        Dictionary of exp data of type rhol.

        * name : str, data type, in this case RhoL
        * calctype : str, Optional, default: 'liquid_properties'
        * T : list, List of temperature values for calculation
        * xi : list, List of liquid mole fractions used in liquid_properties calculations
        * weights : list/float, Either a list as long as the number of data points to multiply by the objective value associated with each point, or a float to multiply the objective value of this data set.
        * rhodict : dict, Optional, default: {"minrhofrac":(1.0 / 60000.0), "rhoinc":10.0, "vspacemax":1.0E-4}, Dictionary of options used in calculating pressure vs. mole fraction curves.

    Attributes
    ----------
    name : str
        Data type, in this case rhol
    calctype : str, Optional, default: 'liquid_properties'
        Thermodynamic calculation type
    T : list
        List of temperature values for calculation
    xi : list
        List of liquid mole fractions, sum(xi) should equal 1
    """

    def __init__(self, data_dict):

        logger = logging.getLogger(__name__)

        # Self interaction parameters
        self.name = data_dict["name"]
        try:
            self.calctype = data_dict["calctype"]
        except:
            self.calctype = "liquid_properties"

        data_type = []
        data_type_name = []
        if "xi" in data_dict:
            self.xi = data_dict["xi"]
            data_type.append(self.xi)
            data_type_name.append("xi")
        if "T" in data_dict:
            self.T = data_dict["T"]
            data_type.append(self.T)
            data_type_name.append("T")
        if "rhol" in data_dict:
            self.rhol = data_dict["rhol"]
            data_type.append(self.rhol)
            data_type_name.append("rhol")
        if "P" in data_dict:
            if (type(P) == float or len(P)==1):
                self.P = np.ones(self.T)*P
            else:
                self.P = P
        else:
            self.P = np.ones(self.T)*101325.0
            logger.info("Assume atmospheric pressure")
        data_type.append(self.P)
        data_type_name.append("P")

        tmp = ["xi","T","rhol"]
        if not all([hasattr(self,x) for x in tmp]):
            raise ImportError("Given liquid property data, values for T, xi, and rhol should have been provided.")

        try:
            self.weights = data_dict["weights"]
        except:
            self.weights = 1.0

        try:
            self._rhodict = data_dict["rhodict"]
        except:
            self._rhodict = {"minrhofrac":(1.0 / 300000.0), "rhoinc":10.0, "vspacemax":1.0E-4}

        logger.info("Data type 'liquid_properties' initiated with calctype, {}, and data types: {}".format(self.calctype,", ".join(data_type_name)))

    def _thermo_wrapper(self, eos):

        """
        Generate thermodynamic predictions from eos object

        Parameters
        ----------
        eos : obj
            EOS object with updated parameters

        Returns
        -------
        phase_list : float
            A list of the predicted thermodynamic values estimated from thermo calculation. This list can be composed of lists or floats
        """

        try:
            output_dict = thermo(eos,{"calculation_type":self.calctype,"Tlist":self.T,"xilist":self.xi,"Plist":self.P,"rhodict":self._rhodict})
            output = output_dict["rhol"]
        except:
            raise ValueError("Calculation of calc_rhol failed")
        return output


    def objective(self, eos):

        """
        Generate objective function value from this dataset

        Parameters
        ----------
        eos : obj
            EOS object with updated parameters

        Returns
        -------
        obj_val : float
            A value for the objective function
        """

        phase_list = self._thermo_wrapper(eos)

        # Reformat array of results
        phase_list, len_list = ff.reformat_ouput(phase_list) 
        phase_list = np.transpose(np.array(phase_list))

        # objective function
        obj_value = np.sum((((phase_list[0] - self.rhol) / self.rhol)**2)*self.weights)

        return obj_value

    def __str__(self):

        string = "Data Set Object\nname: %s\ncalctype:%s\nNdatapts:%g" % {self.name, self.calctype, len(self.T)}
        return string



"""
    This file contains the interface needed to create further objects to represent experimental data. 

    Using this template all future data types will be easily exchanged.
    
"""

# All folders in this directory refer back to this interface

from abc import ABC, abstractmethod


# __________________ EOS Interface _________________
class ExpDataTemplate(ABC):

    """
    Interface needed to create further objects to represent experimental data.

     Using this template all future data types will be easily exchanged.
    """

    @abstractmethod
    def objective(self, eos):
        """
        Float representing objective function of from comparing predictions to experimental data.
        """
        pass




# -- coding: utf8 --

"""
    EOS object for Peng-Robinson
    
"""

import numpy as np
import logging

from despasito.equations_of_state.interface import EOStemplate



class cubic_peng_robinson(EOStemplate):

    """
    EOS object for the Peng-Robinson EOS. 

    All input and calculated parameters are defined as hidden attributes.
    
    Parameters
    ----------
    xi : list[float]
        Mole fraction of component, only relevant for parameter fitting
    beads : list[str]
        List of unique component names
    beadlibrary : dict
        A dictionary where bead names are the keys to access EOS self interaction parameters:

        - Tc: :math:`T_{C}`, Critical temperature [K]
        - Pc: :math:`P_{C}`, Critical pressure [Pa]
        - omega: :math:`\omega`, Acentric factor 

    crosslibrary : dict, Optional, default: {}
        Optional library of bead cross interaction parameters. As many or as few of the desired parameters may be defined for whichever group combinations are desired.

        - kij: :math:`k_{ij}`, binary interaction parameter
        
    Attributes
    ----------
    T : float, default: numpy.nan
        Temperature value is initially defined as NaN for a placeholder until temperature dependent attributes are initialized by using a method of this class.
    
    """

    def __init__(self, kwargs):

        logger = logging.getLogger(__name__)

        # Self interaction parameters
        self._beads = kwargs['beads']
        self._beadlibrary = kwargs['beadlibrary']
        self._nui = np.identity(len(self._beads))

        self._Tc = np.zeros(len(self._beads))
        self._Pc = np.zeros(len(self._beads))
        self._omega = np.zeros(len(self._beads))
        self._kappa = np.zeros(len(self._beads))
        self.alpha = np.empty(len(self._beads))
        self.ai = np.zeros(len(self._beads))
        self.bi = np.zeros(len(self._beads))

        self._R = 8.31446261815324 # [J/mol*K]

        for bead in self._beads:
            if bead in self._beadlibrary:
                ind = self._beads.index(bead)
                try:
                    self._Tc[ind] = self._beadlibrary[bead]["Tc"]
                    self._Pc[ind] = self._beadlibrary[bead]["Pc"]
                    if "omega" in self._beadlibrary[bead]:
                        self._omega[ind] = self._beadlibrary[bead]["omega"]     
                        self._kappa[ind] = 0.37464 + 1.54226*self._omega[ind] - 0.26992*self._omega[ind]**2

                    self.ai[ind] = 0.45723553*(self._R*self._Tc[ind])**2/self._Pc[ind]
                    self.bi[ind] = 0.07779607*(self._R*self._Tc[ind]/self._Pc[ind])

                except:
                    raise ValueError("Either 'Tc' or 'Pc' was not provided for component: {}".format(bead))
            else:
                raise ValueError("Parameters weren't provided for component: {}".format(bead))

        # Cross interaction parameters
        self._kij = np.zeros((len(self._beads),len(self._beads)))
        if 'crosslibrary' in list(kwargs.keys()):
            crosslibrary = kwargs['crosslibrary']
            for key, value in crosslibrary.items():
                if key in self._beads:
                    ind = self._beads.index(key)
                    for key2, value2 in value.items():
                        if key2 in self._beads:
                            jnd = self._beads.index(key2)
                            if "kij" in value2:
                                self._kij[ind,jnd] = value2["kij"]
                                self._kij[jnd,ind] = value2["kij"]
                                logger.info("Parameter 'kij' accepted for interactions between {} and {}".format(key,key2))

        # Initialize temperature attribute
        self.T = np.nan
        self.aij = np.nan
        self.bij = np.nan

    def _calc_temp_dependent_parameters(self,T):
        """
        Compute ai and alpha given temperature
       
        Parameters
        ----------
        T : float
            Temperature of the system [K]

        Attributes
        ----------
        ai : numpy.ndarray
            Peng-Robinson parameter a [m^6/mol^2]
        alpha : numpy.ndarray
            Peng-Robinson parameter b [m^3/mol]
        """
        for bead in self._beads: 
            if bead in self._beadlibrary:
                ind = self._beads.index(bead)
                if self._kappa[ind]:
                    self.alpha[ind] = (1+self._kappa[ind]*(1-np.sqrt(T/self._Tc[ind])))**2
                else:
                    self.alpha[ind] = 1.0

    def _calc_mixed_parameters(self,xi,T):

        """
        Compute mixing aij and bij given composition
       
        Parameters
        ----------
        xi : list[float]
            Mole fraction of each component
        T : float
            Temperature of the system [K]

        Attributes
        ----------
        aij : float
            Peng-Robinson parameter a [m^6/mol^2]
        bij : float
            Peng-Robinson parameter b [m^3/mol]
        """

        if T != self.T:
            self.T = T
            self._calc_temp_dependent_parameters(T)

        aij = 0
        index = range(len(xi))
        for i in index:
            for j in index:
                aij += xi[i]*xi[j]*np.sqrt(self.ai[i]*self.alpha[i]*self.ai[j]*self.alpha[j])*(1.-self._kij[i][j])

        self.aij = aij
        self.bij = np.sum(xi*self.bi)

    def P(self, rho, T, xi):
        """
        Compute pressure given system information
       
        Parameters
        ----------
        rho : numpy.ndarray
            Number density of system [mol/m^3]
        T : float
            Temperature of the system [K]
        xi : list[float]
            Mole fraction of each component
       
        Returns
        -------
        P : numpy.ndarray
            Array of pressure values [Pa] associated with each density and so equal in length
        """

        #logger = logging.getLogger(__name__)

        if T != self.T:
            self.T = T
            self._calc_temp_dependent_parameters(T)

        self._calc_mixed_parameters(xi,T)
        
        if np.isscalar(rho):
            rho = np.array([rho])
        elif type(rho) != np.ndarray:
            rho = np.array(rho)

        P = self._R*self.T * rho / (1-self.bij*rho) - rho**2*self.aij/((1+self.bij*rho)+rho*self.bij*(1-self.bij*rho))

        return P

    def fugacity_coefficient(self, P, rho, xi, T):

        """
        Compute fugacity coefficient
      
        Parameters
        ----------
        P : float
            Pressure of the system [Pa]
        rho : float
            Molar density of system [mol/m^3]
        T : float
            Temperature of the system [K]
        xi : list[float]
            Mole fraction of each component
    
        Returns
        -------
        mui : numpy.ndarray
            :math:`\mu_i`, Array of chemical potential values for each component
        """

        #logger = logging.getLogger(__name__)

        if T != self.T:
            self.T = T
            self._calc_temp_dependent_parameters(T)

        self._calc_mixed_parameters(xi,T)

        Z = P/(T*self._R*rho)
        Ai = self.ai*self.alpha*P/(self._R*T)**2
        Bi = self.bi*P/(self._R*T)
        B = self.bij*P/(self._R*T)
        A = self.aij*P/(self._R*T)**2

        sqrt2 = np.sqrt(2.0)
        tmp1 = A/(2.0*sqrt2*B)*np.log((Z+(1+sqrt2)*B)/(Z+(1-sqrt2)*B))
        tmp3 = Bi*(Z-1)/B-np.log(Z-B)
        tmp2 = np.zeros(len(xi))

        index = range(len(xi))
        for i in index:
            Aij = np.zeros(len(xi))
            for j in index:
                Aij[j] = np.sqrt(Ai[i]*Ai[j])*(1.-self._kij[i][j])
            tmp2[i] = Bi[i]/B - 2*np.sum(xi*Aij)/A
        phi = np.exp(tmp1*tmp2+tmp3)

        return phi

    def density_max(self, xi, T, maxpack=0.9):

        """
        Estimate the maximum density based on the hard sphere packing fraction.
        
        Parameters
        ----------
        xi : list[float]
            Mole fraction of each component
        T : float
            Temperature of the system [K]
        maxpack : float, Optional, default: 0.65
            Maximum packing fraction
        
        Returns
        -------
        maxrho : float
            Maximum molar density [mol/m^3]
        """

        #logger = logging.getLogger(__name__)

        if T != self.T:
            self.T = T
            self._calc_temp_dependent_parameters(T)

        self._calc_mixed_parameters(xi,T)

        maxrho = maxpack /self.bij

        return maxrho

    def param_guess(self,fit_params):
        """
        Generate initial guesses for the parameters to be fit.

        Parameters
        ----------
        fit_params : list[str]
            A list of parameters to be fit. See EOS documentation for supported parameter names. Cross interaction parameter names should be composed of parameter name and the other bead type, separated by an underscore (e.g. epsilon_CO2).

        Returns
        -------
        param_initial_guesses : numpy.ndarray, 
            An array of initial guesses for parameters, these will be optimized throughout the process.
    """

        #logger = logging.getLogger(__name__)

        l_fitparams = len(fit_params)
        if l_fitparams == 1:
            param_initial_guesses = np.array([l_fitparams])
        else:
            param_initial_guesses = np.zeros(1,l_fitparams)

        for i,param in enumerate(fit_params):
            if param == "ai":
                param_initial_guesses[i] = 1e+7
            elif param == "bi":
                param_initial_guesses[i] = 1e+11
            else:
                raise ValueError("The parameter name %s does not fall under any of the categories, ai or bi") 

        return param_initial_guesses

    def update_parameters(self, param_name, bead_names, param_value):
        r"""
        Update a single parameter value to _beadlibrary or _crosslibrary attributes during parameter fitting process. 

        To refresh those parameters that are dependent on these libraries, use method "parameter refresh".
        
        Parameters
        ----------
        param_name : str
            name of parameters being updated
        bead_names : list
            List of bead names to be changed. For a self interaction parameter, the length will be 1, for a cross interaction parameter, the length will be two.
        param_value : float
            Value of parameter
        """

        #logger = logging.getLogger(__name__)

        param_types = ["ai", "bi", "kij"]

        if len(bead_names) > 2:
            raise ValueError("The bead names {} were given, but only a maximum of 2 are permitted.".format(", ".join(bead_names)))
        if not set(bead_names).issubset(self._beads):
            raise ValueError("The bead names {} were given, but they are not in the allowed list: {}".format(", ".join(bead_names),", ".join(self._beads)))

        # Non bonded parameters
        if (param_name in param_types):
            # Parameter kij
            if "kij" == param_name:
                if bead_names[0] in self._beads:
                    ind = self._beads.index(bead_names[0])
                    if bead_names[1] in self._beads:
                        jnd = self._beads.index(bead_names[1])
                        self._kij[ind,jnd] = param_value
                        self._kij[jnd,ind] = param_value
            elif "ai" == param_name:
                if bead_names[0] in self._beads:
                    ind = self._beads.index(bead_names[0])
                    self.ai[ind] = param_value
            elif "bi" == param_name:
                if bead_names[0] in self._beads:
                    ind = self._beads.index(bead_names[0])
                    self.bi[ind] = param_value

        else:
            raise ValueError("The parameter name %s is not found in the allowed parameter types: %s" % (param_name,", ".join(param_types)))

    def __str__(self):

        string = "Beads:" + str(self._beads) + "\n"
        string += "T:" + str(self.T) + "\n"
        return string





import numpy as np
cimport numpy as np
import logging
import os

from .constants import ckl_coef

def calc_a1s(rho, Cmol2seg, l_kl, zetax, epsilonkl, dkl):
    r""" wrapper function for calling 2d/3d versions of calc_a1s ... this is done for stupid Numba 
    """
    if len(l_kl.shape) == 2:
        output = calc_a1s_2d(rho, Cmol2seg, l_kl, zetax, epsilonkl, dkl)
    elif len(l_kl.shape) == 1:
        output = calc_a1s_1d(rho, Cmol2seg, l_kl, zetax, epsilonkl, dkl)

    return output

def calc_a1s_2d(rho, Cmol2seg, l_kl, zetax, epsilonkl, dkl):
    r""" 
    Return a1s,kl(rho*Cmol2seg,l_kl) in K as defined in eq. 25, used in the calculation of :math:`A_1` the first order term of the perturbation expansion corresponding to the mean-attractive energy.

    Parameters
    ----------
    rho : numpy.ndarray
        Number density of system [molecules/m^3]
    Cmol2seg : float
        Conversion factor from from molecular number density, :math:`\rho`, to segment (i.e. group) number density, :math:`\rho_S`. Shown in eq. 13
    l_kl : numpy.ndarray
        Matrix of mie potential exponents for k,l groups
    zetax : numpy.ndarray
        Matrix of hypothetical packing fraction based on hard sphere diameter for groups (k,l)
    epsilonkl : numpy.ndarray
        Matrix of well depths for groups (k,l)
    dkl : numpy.ndarray
        Matrix of hardsphere diameters for groups (k,l)

    Returns
    -------
    numpy.ndarray
        Matrix used in the calculation of :math:`A_1` the first order term of the perturbation expansion corresponding to the mean-attractive energy, size is the Ngroups by Ngroups

    :note: output seems to be a tensor of size (N x Ngroups x Ngroups)
    """
    # Andrew: why is the 4 hard-coded here?
    nbeads = len(dkl)
    zetax_pow = np.empty((len(rho), 4), dtype=rho.dtype)
    zetax_pow[:, 0] = zetax
    for i in range(1,4):
        zetax_pow[:, i] = zetax_pow[:, i-1] * zetax_pow[:, 0]

    # check if you have more than 1 bead types
    etakl = np.empty((len(rho), nbeads, nbeads), dtype=rho.dtype)

    for k in range(nbeads):
        for l in range(nbeads):
            tmp = np.dot(ckl_coef, np.array( (1.0, 1.0/l_kl[k, l], 1.0/l_kl[k, l]**2, 1.0/l_kl[k, l]**3), dtype=ckl_coef.dtype ))
            etakl[:, k, l] = np.dot( zetax_pow, tmp )

    a1s = - (1.0 - (etakl / 2.0)) / ((1.0 - etakl)**3) * 2.0 * np.pi * Cmol2seg * ((epsilonkl * (dkl**3)) / (l_kl - 3.0))
    return np.transpose(np.transpose(a1s) * rho)

def calc_a1s_1d(rho, Cmol2seg, l_kl, zetax, epsilonkl, dkl):
    r""" 
    Return a1s,kl(rho*Cmol2seg,l_kl) in K as defined in eq. 25, used in the calculation of :math:`A_1` the first order term of the perturbation expansion corresponding to the mean-attractive energy.

    Parameters
    ----------
    rho : numpy.ndarray
        Number density of system [molecules/m^3]
    Cmol2seg : float
        Conversion factor from from molecular number density, :math:`\rho`, to segment (i.e. group) number density, :math:`\rho_S`. Shown in eq. 13
    l_kl : numpy.ndarray
        Matrix of mie potential exponents for k,l groups
    zetax : numpy.ndarray
        Matrix of hypothetical packing fraction based on hard sphere diameter for groups (k,l)
    epsilonkl : numpy.ndarray
        Matrix of well depths for groups (k,l)
    dkl : numpy.ndarray
        Matrix of hardsphere diameters for groups (k,l)

    Returns
    -------
    numpy.ndarray
        Matrix used in the calculation of :math:`A_1` the first order term of the perturbation expansion corresponding to the mean-attractive energy, size is the Ngroups by Ngroups
    """
    nbeads = len(dkl)
    zetax_pow = np.empty((len(rho), 4), dtype=rho.dtype)
    zetax_pow[:, 0] = zetax
    for i in range(1,4):
        zetax_pow[:, i] = zetax_pow[:, i-1] * zetax_pow[:, 0]

    # check if you have more than 1 bead types
    etakl = np.empty((len(rho), nbeads), dtype=rho.dtype)

    for k in range(nbeads):
        tmp = np.dot(ckl_coef, np.array( (1.0, 1.0/l_kl[k], 1.0/l_kl[k]**2, 1.0/l_kl[k]**3), dtype=ckl_coef.dtype ) )
        etakl[:, k] = np.dot( zetax_pow, tmp )

    a1s = - (1.0 - (etakl / 2.0)) / (1.0 - etakl)**3 * 2.0 * np.pi * Cmol2seg * ((epsilonkl * (dkl**3)) / (l_kl - 3.0) )

    return np.transpose(np.transpose(a1s) * rho)



"""
Thermodynamics

This package will take in an equation of state object, and any user defined variables for calculation. The calculation type will then be compared to those available in the thermo.py file and be executed.

"""

# Add imports here
from inspect import getmembers, isfunction
#import logging

from . import calc_types

def thermo(eos, thermo_dict):
    """
    Use factory design pattern to search for matching calctype with those supported in this module.
    
    To add a new calculation type, add a function to thermo.py in the thermodynamics module..

    Parameters
    ----------
        eos : obj
            Equation of state output that writes pressure, max density, and chemical potential
        thermo_dict : dict
            Other keywords passed to the function, depends on calculation type
                

    Returns
    -------
        output_dict : dict
            Output of dictionary containing given and calculated values
    """

    #logger = logging.getLogger(__name__)

    try:
        calctype = thermo_dict['calculation_type']
    except:
        raise Exception('No calculation type specified')

    # Extract available calculation types
    calc_list = [o[0] for o in getmembers(calc_types) if isfunction(o[1])]

    # Unpack inputs and check
    sys_dict, kwargs = {}, {}
    for key, value in thermo_dict.items():
        if key not in ['output_file','calculation_type']:
            sys_dict[key] = value
        elif key != 'calculation_type':
            kwargs[key] = value

    try:
        func = getattr(calc_types, calctype)
    except:
        raise ImportError("The calculation type, '"+calctype+"', was not found\nThe following calculation types are supported: "+", ".join(calc_list))

    try:
        output_dict = func(eos, sys_dict, **kwargs)
    except:
        raise TypeError("The calculation type, '"+calctype+"', failed")

    return output_dict




"""
Regression tests for the despasito input/output package.

The following functions are not tested because they require importing a file:
    file2paramdict, extract_calc_data, process_commandline, make_xi_matrix

The following functions are not tested because they require exporting a file:
    write_EOSparameters

The following functions are tested:
    process_bead_data
"""

# Import package, test suite, and other packages as needed
import despasito.input_output.readwrite_input as d_io
import pytest
import numpy as np

# Not Used because we shouldn't reference an external file
@pytest.mark.parametrize('data', [([[['CH4_2', 1]], [['eCH3', 2]]])])
def test_process_bead_data(data):
    """Test extraction of system component information"""

    beads, nui = d_io.process_bead_data(data)

    errors = []
    if not set(beads) == set(['CH4_2', 'eCH3']):
        errors.append("beads: %s is not %s" % (str(beads),str(['CH4_2', 'eCH3'])))
    elif not np.array_equal(nui,np.array([[1.,0.],[0.,2.]])):
        errors.append("nui: %s is not %s" % (str(nui),str(np.array([[1.,0.],[0.,2.]]))))

    assert not errors, "errors occured:\n{}".format("\n".join(errors))

# Not Used because we shouldn't reference an external file
#@pytest.mark.parametrize('key, answer', [("rhoinc",2.0),("minrhofrac",2.5e-06)])
#def test_file2paramdict(key,answer):
#    """Test conversion of txt file to dictionary"""
#    rho_dict = d_io.file2paramdict("example/dens_params.txt")
#    print(rho_dict)
#    assert rho_dict[key] == pytest.approx(answer,abs=1e-7)



"""
Authors: Andrew Abi-Mansour <andrew.gaam [at] gmail [dot] com>
         Pierre Ablin <pierreablin [at] gmail [dt] com>

License: MIT

"""
import autograd.numpy as np
import autograd
from scipy import optimize

class Minimizer:
    """A wrapper class for scipy.optimize.minimize that computes derivatives with JAX (AD).
        Parameters
        ----------
        objective_function : callable
            The objective function to be minimized.
                ``fun(optim_vars, *args) -> float``
            or
                ``fun(*optim_vars, *args) -> float``
            where optim_vars is either a numpy array or a list of numpy
            arrays and `args` is a tuple of fixed parameters needed to
            completely specify the function.
        optim_vars : ndarray or list of ndarrays
            Initial guess.
        args : tuple, optional
            Extra arguments passed to the objective function.
        precon_fwd : callable, optional
            The forward preconditioning.
                ``fun(optim_vars, *args) -> precon_optim_vars``
            or
                ``fun(*optim_vars, *args) -> precon_optim_vars``
            where optim_vars is either a numpy array or a list of numpy
            arrays and `args` is a tuple of fixed parameters needed to
            completely specify the function.
            The optimized function will be the composition:
            `objective_function(precon_fwd(optim_vars))`.
        precon_bwd : callable, optional
            The backward preconditioning.
                ``fun(precon_optim_vars, *args) -> optim_vars``
            or
                ``fun(*precon_optim_vars, *args) -> optim_vars``
            where optim_vars is either a numpy array or a list of numpy
            arrays and `args` is a tuple of fixed parameters needed to
            completely specify the function.
            This should be the reciprocal function of precon_fwd.
        kwargs : dict, optional
            Extra arguments passed to scipy.optimize.minimize. See
            https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html
            for the full list of available keywords.
        Returns
        -------
        output : ndarray or list of ndarrays
            The solution, of same shape as the input `optim_vars`.
        res : OptimizeResult
            The optimization result represented as a ``OptimizeResult`` object.
            Important attributes are: ``x`` the solution array, ``success`` a
            Boolean flag indicating if the optimizer exited successfully and
            ``message`` which describes the cause of the termination. See
            `OptimizeResult` for a description of other attributes.
        """

    def __init__(self, objective_function, optim_vars, args=(), precon_fwd=None,
             precon_bwd=None, **kwargs):

        # Check if there is preconditioning:
        self._precondition = precon_fwd is not None
        if self._precondition != (precon_bwd is not None):

            error_string = {True: 'precon_fwd', False: 'precon_bwd'}[self._precondition]
            raise ValueError(f'You should specify both precon_fwd and precon_bwd,'
                             ' you only specified {error_string}')

        self._optim_vars = optim_vars
        self._objective_function = objective_function
        self._args = args
        self._precon_fwd = precon_fwd
        self._precon_bwd = precon_bwd
        self._kwargs = kwargs

    @staticmethod
    def _vectorize(optim_vars):
        shapes = [var.shape for var in optim_vars]
        x = np.concatenate([var.ravel() for var in optim_vars])
        return x, shapes

    @staticmethod
    def _split(x, shapes):
        x_split = np.split(x, np.cumsum([np.prod(shape) for shape in shapes[:-1]]))
        optim_vars = [var.reshape(*shape) for (var, shape) in zip(x_split, shapes)]
        return optim_vars

    def _objFunc(self, x):
        def _scipy_func(objective_function, gradient, x, shapes, args=()):
            """ whatever ... """
            optim_vars = Minimizer._split(x, shapes)
            obj = objective_function(optim_vars, *args)
            gradients = gradient(optim_vars, *args)
            g_vectorized, _ = Minimizer._vectorize(gradients)
        
            return obj, g_vectorized

        print(x)
        return _scipy_func(self.objective_converted, self._gradient, x, self._shapes, self._args)

    def minimize(self, **kwargs):

        if 'objective_function' in kwargs:
            self._objective_function = kwargs['objective_function']

        if 'optim_vars' in kwargs:
            self._optim_vars = kwargs['optim_vars']

        if 'args' in kwargs:
            self._args = kwargs['args']

        if self._precondition:

            self._optim_vars = self._convert_to_tuple(self._optim_vars)
            self._args = args
            precon_optim_vars = precon_fwd(*self._optim_vars, *args)

            precon_result, res = self.minimize(
                objective_function=self.precon_objective, 
                optim_vars = precon_optim_vars,
                args=args, precon_fwd=None, precon_bwd=None, 
                **kwargs)

            precon_result = self._convert_to_tuple(precon_result)
            return precon_bwd(*precon_result, *args), res

        # Check if there are bounds:
        bounds = self._kwargs.get('bounds')
        bounds_in_kwargs = bounds is not None

        # Convert input to a list if it is a single array
        if type(self._optim_vars) is np.ndarray:
            input_is_array = True
            self._optim_vars = (self._optim_vars,)
            if bounds_in_kwargs:
                bounds = (bounds,)
        else:
            input_is_array = False

        # Compute the gradient
        self._gradient = autograd.grad(self.objective_converted)

        # Vectorize optimization variables
        x0, self._shapes = Minimizer._vectorize(self._optim_vars)

        # Convert bounds to the correct format
        if bounds_in_kwargs:
            bounds = self._convert_bounds(bounds, shapes)
            self._kwargs['bounds'] = bounds

        res = optimize.minimize(self._objFunc, x0, jac=True, **self._kwargs)

        # Convert output to the input format
        output = Minimizer._split(res['x'], self._shapes)
        if input_is_array:
            output = output[0]

        return output, res

    def precon_objective(self, *precon_optim_vars_and_args):
        """ whatever ... """

        args = precon_optim_vars_and_args[-len(self._args):]
        optim_vars = precon_bwd(*precon_optim_vars_and_args)
        optim_vars = self._convert_to_tuple(optim_vars)
            
        return objective_function(*optim_vars, *args)

    def _convert_to_tuple(self, optim_vars):
        if type(optim_vars) not in (list, tuple):
            return (optim_vars,)
    
        return optim_vars

    def objective_converted(self, optim_vars, *args):
        """ Converts loss to readable jax format """
        return self._objective_function(*optim_vars, *args)

    def _convert_bounds(self, bounds, shapes):
        output_bounds = []
        for shape, bound in zip(shapes, bounds):
            # Check is the bound is already parsable by scipy.optimize
            b = bound[0]
            if isinstance(b, (list, tuple, np.ndarray)):
                output_bounds += bound
            else:
                output_bounds += [bound, ] * np.prod(shape)
        return output_bounds


# Performance test function
def objFunc(x):
    return optimize.rosen(x)



if __name__ == '__main__':
    
    x0 = np.array([1.3, 0.7, 0.8, 1.9, 1.2])
 
    Min = Minimizer(objFunc, x0, options={'disp': False})
    x_min, _ = Min.minimize()
    
    x_min_scipy = optimize.minimize(objFunc, x0, method='nelder-mead',
                options={'disp': False})

    print('output:', x_min, x_min_scipy.x)



"""
Fit Parameters
--------------

This package uses functions from input_output, equations_of_state, and thermodynamics to fit parameters to experimental data.

"""

import os
import numpy as np
from importlib import import_module
import scipy.optimize as spo
import logging

from . import fit_funcs as ff
from . import data_classes

def fit(eos, thermo_dict):
    r"""
    Fit defined parameters for equation of state object with given experimental data. 

    Each set of experimental data is converted to an object with the built in ability to evaluate its part of objective function.
    To add another type of supported experimental data, add a class to the fit_classes.py file.

    Parameters
    ----------
    eos : obj
        Equation of state output that writes pressure, max density, chemical potential, updates parameters, and evaluates objective functions. For parameter fitting algorithm See equation of state documentation for more details.
    thermo_dict : dict
        Dictionary of instructions for thermodynamic calculations and parameter fitting.

        - opt_params (dict) - Parameters used in basin fitting algorithm.

            - fit_bead (str) - Name of bead whose parameters are being fit, should be in bead list of beadconfig
            - fit_params (list[str]) - This list of contains the name of the parameter being fit (e.g. epsilon). See EOS documentation for supported parameter names. Cross interaction parameter names should be composed of parameter name and the other bead type, separated by an underscore (e.g. epsilon_CO2).
            - beadparams0 (list[float]), Optional - Initial guess in parameter. If one is not provided, a guess is made based on the type of parameter from eos object.

        - bounds (numpy.ndarray) - List of length equal to fit_params with lists of pairs for minimum and maximum bounds of parameter being fit. Defaults are broad, recommend specification.
        - exp_data (dict) - This dictionary is made up of a dictionary for each data set that the parameters are fit to. Each dictionary is converted into an object and saved back to this structure before parameter fitting begins. Each key is an arbitrary string used to identify the data set and used later in reporting objective function values during the fitting process. See data type objects for more details.

            - name (str) - One of the supported data type objects to fit parameters

        - basin_dict (dict), Optional - kwargs used in scipy.optimize.basinhopping

            - niter (int) - Number of basin hopping iterations
            - T (float) - Temperature parameter, should be comparable to separation between local minima (i.e. the “height” of the walls separating values).
            - take_step (obj) - Callable object that includes custom routine that produces attribute stepsize
            - niter_success (int) - Stop run if minimum stays the same for this many iterations

        - minimizer_dict (dict), Optional - Dictionary used to define minimization type and the associated options.

            - method (str) - Method available to scipy.optimize.minimize
            - options (dict) - This dictionary contains the kwargs available to the chosen method
  
    Returns
    -------
    Output file saved in current working directory
    """

    logger = logging.getLogger(__name__)

    # Extract relevant quantities from thermo_dict
    keys_del = []
    for key, value in thermo_dict.items():
        # Extract inputs
        if key == "opt_params":
            opt_params  = value
        elif key == "exp_data":
            exp_data = value
        # Optional inputs
        elif key == "bounds":
            bounds = value
        elif key == "minimizer_dict":
            minimizer_dict = value
        elif key == "basin_dict":
            basin_dict = value
        else:
            continue
        keys_del.append(key)

    for key in keys_del:
        thermo_dict.pop(key,None)

    if list(thermo_dict.keys()):
        logger.info("Note: thermo_dict keys: {}, were not used.".format(", ".join(list(thermo_dict.keys()))))    

    # Reformat exp. data into formatted dictionary
    exp_dict = {}
    pkgpath = os.path.dirname(data_classes.__file__)
    type_list = [f for f in os.listdir(pkgpath) if ".py" in f]
    type_list = type_list.remove("__init__.py")

    for key, data_dict in exp_data.items():
        fittype = data_dict["name"]
        try:
            exp_module = import_module("."+fittype,package="despasito.fit_parameters.data_classes")
            data_class = getattr(exp_module, "Data")
        except:
            if not type_list: raise ImportError("No fit types")
            elif len(type_list) == 1: tmp = type_list[0]
            else: tmp = ", ".join(type_list)
            raise ImportError("The experimental data type, '"+fittype+"', was not found\nThe following calculation types are supported: "+tmp)

        try:
            instance = data_class(data_dict)
            exp_dict[key] = instance
            logger.info("Initiated exp. data object: {}".format(instance.name))
        except:
            raise AttributeError("Data set, {}, did not properly initiate object".format(key))

    # Generate initial guess for parameters if none was given
    if "beadparams0" in opt_params:
        beadparams0 = opt_params["beadparams0"]
        logger.info("Initial guess in parameters provided: {}".format(beadparams0))
    else:
        beadparams0 = eos.param_guess(opt_params["fit_params"])

    # Options for basin hopping
    new_basin_dict = {"niter": 10, "T": 0.5, "niter_success": 3}
    if "basin_dict" in thermo_dict:
        for key, value in basin_dict.items():
            new_basin_dict[key] = value
    basin_dict = new_basin_dict

    # Set up options for minimizer in basin hopping
    new_minimizer_dict = {"method": 'nelder-mead', "options": {'maxiter': 50}}
    if "minimizer_dict" in thermo_dict:
        for key, value in minimizer_dict.items():
            if key == "method":
                new_minimizer_dict[key] = value
            elif key == "options":
                for key2, value2 in value.items():
                    new_minimizer_dict[key][key2] = value2
    minimizer_dict = new_minimizer_dict
  
    # NoteHere: how is this array generated? stepmag = np.array([550.0, 26.0, 4.0e-10, 0.45, 500.0, 150.0e-30, 550.0])
    try:
        if "stepsize" in basin_dict:
        	stepsize = basin_dict["stepsize"]
        else:
        	stepsize = 0.1
        stepmag = np.transpose(np.array(bounds))[1]
        basin_dict["take_step"] = ff.BasinStep(stepmag, stepsize=stepsize)
        custombounds = ff.BasinBounds(bounds)
    except:
    	raise TypeError("Could not initialize BasinStep and/or BasinBounds")

    # Run Parameter Fitting
    try:
        result = spo.basinhopping(ff.compute_SAFT_obj, beadparams0, **basin_dict, accept_test=custombounds, disp=True,
                       minimizer_kwargs={"args": (opt_params, eos, exp_dict),**minimizer_dict})

        logger.info("Fitting terminated:\n{}".format(result.message))
        logger.info("Best Fit Parameters")
        logger.info("    Obj. Value: {}".format(result.fun))
        for i in range(len(opt_params["fit_params"])):
            logger.info("    {} {}: {}".format(opt_params["fit_bead"],opt_params["fit_params"][i],result.x[i]))

    except:
        raise TypeError("The parameter fitting failed")

    return {"fit_bead": opt_params["fit_bead"], "fit_parameters":opt_params["fit_params"], "final_parameters": result.x, "objective_value": result.fun} 




"""
    Interface needed to create further equation of state (EOS) objects. 

    All folders in this directory refer back to this interface. Using this template all future EOS will be easily exchanged.
    
"""

# All folders in this directory refer back to this interface

from abc import ABC, abstractmethod


# __________________ EOS Interface _________________
class EOStemplate(ABC):

    """
    Interface used in all EOS object options.

    By using this template, all EOS objects are then easily exchanged.
    """

    @abstractmethod
    def P(self, rho, T, xi):
        """
        Output pressure value predicted by EOS.
        """
        pass

    @abstractmethod
    def fugacity_coefficient(self, P, rho, xi, T):
        """
        Output chemical potential predicted by EOS.
        """
        pass

    @abstractmethod
    def density_max(self, xi, T):
        """
        Output maximum packing density predicted by EOS.
        """
        pass

    @abstractmethod
    def param_guess(self, fit_params):
        """
        Output a guess for the given parameter type.
        """
        pass

    @abstractmethod
    def update_parameters(self, param_name, bead_names, param_value):
        """
        Update a given parameter in EOS.
        """
        pass

